@@ -4,13 +4,10 @@
 
 import numpy as np
 import pandas as pd
-import random
-import timeit
-import copy
+from tqdm import tqdm
+
 from collections.abc import Iterable
 from sklearn.neighbors import KDTree
-
-import dice_ml.diverse_counterfactuals as exp
 from dice_ml.counterfactual_explanations import CounterfactualExplanations
 from dice_ml.utils.exception import UserConfigValidationException
 
@@ -51,7 +48,7 @@ class ExplainerBase:
             permitted_range=None, features_to_vary="all",
             stopping_threshold=0.5, posthoc_sparsity_param=0.1,
             posthoc_sparsity_algorithm="linear", verbose=False, **kwargs):
-        """Generate counterfactuals by randomly sampling features.
+        """General method for generating counterfactuals.
 
         :param query_instances: Input point(s) for which counterfactuals are to be generated. This can be a dataframe with one or more rows.
         :param total_CFs: Total number of counterfactuals required.
@@ -71,7 +68,8 @@ class ExplainerBase:
         :returns: A CounterfactualExplanations object that contains the list of
         counterfactual examples per query_instance as one of its attributes.
         """
-
+        if total_CFs <= 0:
+            raise UserConfigValidationException("The number of counterfactuals generated per query instance (total_CFs) should be a positive integer.")
         cf_examples_arr = []
         query_instances_list = []
         if isinstance(query_instances, pd.DataFrame):
@@ -79,8 +77,7 @@ class ExplainerBase:
                 query_instances_list.append(query_instances[ix:(ix+1)])
         elif isinstance(query_instances, Iterable):
             query_instances_list = query_instances
-            #query_instances = query_instances.to_dict("records")
-        for query_instance in query_instances_list:
+        for query_instance in tqdm(query_instances_list):
             res = self._generate_counterfactuals(query_instance, total_CFs,
                     desired_class=desired_class,
                     desired_range=desired_range,
@@ -272,13 +269,13 @@ class ExplainerBase:
                 df = cf_examples.final_cfs_df
             for index, row in df.iterrows():
                 for col in self.data_interface.continuous_feature_names:
-                    if not np.isclose(org_instance[col].iloc[0], row[col]):
+                    if not np.isclose(org_instance[col].iat[0], row[col]):
                         if summary_importance is not None:
                             summary_importance[col] += 1
                         if local_importances is not None:
                             local_importances[i][col] += 1
                 for col in self.data_interface.categorical_feature_names:
-                    if org_instance[col].iloc[0] != row[col]:
+                    if org_instance[col].iat[0] != row[col]:
                         if summary_importance is not None:
                             summary_importance[col] += 1
                         if local_importances is not None:
@@ -309,6 +306,9 @@ class ExplainerBase:
         """
         # :param total_random_inits: total random initialization required for algorithm = RandomInitCF (where k CFs are generated by running a CF generation algorithm k times with random initializations.)
 
+        if final_cfs_sparse is None:
+            return final_cfs_sparse
+
         # resetting index to make sure .loc works
         final_cfs_sparse = final_cfs_sparse.reset_index(drop=True)
         # quantiles of the deviation from median for every continuous feature
@@ -328,12 +328,14 @@ class ExplainerBase:
         decimal_prec = dict(zip(self.data_interface.continuous_feature_names, precs))
 
         cfs_preds_sparse = []
+
         for cf_ix in range(len(final_cfs_sparse)):
-            current_pred = self.predict_fn_for_sparsity(final_cfs_sparse.iloc[[cf_ix]][self.data_interface.feature_names])
+            current_pred = self.predict_fn_for_sparsity(
+                final_cfs_sparse.iloc[[cf_ix]][self.data_interface.feature_names])
             for feature in features_sorted:
-                #current_pred = self.predict_fn_for_sparsity(final_cfs_sparse.iloc[[cf_ix]][self.data_interface.feature_names])
+                #current_pred = self.predict_fn_for_sparsity(final_cfs_sparse.iat[[cf_ix]][self.data_interface.feature_names])
                 #feat_ix = self.data_interface.continuous_feature_names.index(feature)
-                diff = query_instance[feature].iloc[0] - final_cfs_sparse.iloc[cf_ix][feature]
+                diff = query_instance[feature].iat[0] - final_cfs_sparse[feature].iat[cf_ix]
                 if(abs(diff) <= quantiles[feature]):
                     if posthoc_sparsity_algorithm == "linear":
                         final_cfs_sparse = self.do_linear_search(diff, decimal_prec, query_instance, cf_ix, feature, final_cfs_sparse, current_pred)
@@ -342,7 +344,6 @@ class ExplainerBase:
                         final_cfs_sparse = self.do_binary_search(diff, decimal_prec, query_instance, cf_ix, feature, final_cfs_sparse, current_pred)
 
             temp_preds = self.predict_fn_for_sparsity(final_cfs_sparse.iloc[[cf_ix]][self.data_interface.feature_names])
-            #temp_preds = np.round(temp_preds.flatten().tolist(), 3)[0]
             cfs_preds_sparse.append(temp_preds)
 
         final_cfs_sparse[self.data_interface.outcome_name] = self.get_model_output_from_scores(cfs_preds_sparse)
@@ -357,25 +358,25 @@ class ExplainerBase:
         current_pred = current_pred_orig
         if self.model.model_type == 'classifier':
             while((abs(diff)>10e-4) and (np.sign(diff*old_diff) > 0) and self.is_cf_valid(current_pred)):
-                old_val = final_cfs_sparse.iloc[cf_ix][feature]
+                old_val = final_cfs_sparse[feature].iat[cf_ix]
                 final_cfs_sparse.loc[cf_ix, feature] += np.sign(diff)*change
                 current_pred = self.predict_fn_for_sparsity(final_cfs_sparse.iloc[[cf_ix]][self.data_interface.feature_names])
                 old_diff = diff
 
                 if not self.is_cf_valid(current_pred):
                     final_cfs_sparse.loc[cf_ix, feature] = old_val
-                    diff = query_instance[feature].iloc[0] - final_cfs_sparse.iloc[cf_ix][feature]
+                    diff = query_instance[feature].iat[0] - final_cfs_sparse[feature].iat[cf_ix]
                     return final_cfs_sparse
 
-                diff = query_instance[feature].iloc[0] - final_cfs_sparse.iloc[cf_ix][feature]
+                diff = query_instance[feature].iat[0] - final_cfs_sparse[feature].iat[cf_ix]
 
         return final_cfs_sparse
 
     def do_binary_search(self, diff, decimal_prec, query_instance, cf_ix, feature, final_cfs_sparse, current_pred):
         """Performs a binary search between continuous features of a CF and corresponding values in query_instance until the prediction class changes."""
 
-        old_val = final_cfs_sparse.iloc[cf_ix][feature]
-        final_cfs_sparse.loc[cf_ix, feature] = query_instance[feature].iloc[0]
+        old_val = final_cfs_sparse[feature].iat[cf_ix]
+        final_cfs_sparse.loc[cf_ix, feature] = query_instance[feature].iat[0]
         # Prediction of the query instance
         current_pred = self.predict_fn_for_sparsity(final_cfs_sparse.iloc[[cf_ix]][self.data_interface.feature_names])
 
@@ -387,8 +388,8 @@ class ExplainerBase:
 
         # move the CF values towards the query_instance
         if diff > 0:
-            left = final_cfs_sparse.iloc[cf_ix][feature]
-            right = query_instance[feature].iloc[0]
+            left = final_cfs_sparse[feature].iat[cf_ix]
+            right = query_instance[feature].iat[0]
 
             while left <= right:
                 current_val = left + ((right - left)/2)
@@ -406,8 +407,8 @@ class ExplainerBase:
                     right = current_val - (10 ** -decimal_prec[feature])
 
         else:
-            left = query_instance[feature].iloc[0]
-            right = final_cfs_sparse.iloc[cf_ix][feature]
+            left = query_instance[feature].iat[0]
+            right = final_cfs_sparse[feature].iat[cf_ix]
 
             while right >= left:
                 current_val = right - ((right - left)/2)
@@ -426,6 +427,22 @@ class ExplainerBase:
 
         return final_cfs_sparse
 
+    def misc_init(self, stopping_threshold, desired_class, desired_range, test_pred):
+        self.stopping_threshold = stopping_threshold
+        if self.model.model_type == 'classifier':
+            self.target_cf_class = np.array(
+                [[self.infer_target_cfs_class(desired_class, test_pred, self.num_output_nodes)]],
+                dtype=np.float32)
+            desired_class = self.target_cf_class[0][0]
+            if self.target_cf_class == 0 and self.stopping_threshold > 0.5:
+                self.stopping_threshold = 0.25
+            elif self.target_cf_class == 1 and self.stopping_threshold < 0.5:
+                self.stopping_threshold = 0.75
+
+        elif self.model.model_type == 'regressor':
+            self.target_cf_range = self.infer_target_cfs_range(desired_range)
+        return desired_class
+
     def infer_target_cfs_class(self, desired_class_input, original_pred,
             num_output_nodes):
         """ Infer the target class for generating CFs. Only called when
@@ -433,21 +450,22 @@ class ExplainerBase:
         TODO: Add support for opposite desired class in multiclass. Downstream methods should decide
         whether it is allowed or not.
         """
-        target_class = None
         if desired_class_input == "opposite":
             if num_output_nodes == 2:
                 original_pred_1 = np.argmax(original_pred)
                 target_class = int(1 - original_pred_1)
-
+                return target_class
             elif num_output_nodes > 2:
-                raise ValueError("Desired class cannot be opposite if the number of classes is more than 2.")
-
-        if isinstance(desired_class_input, int):
+                raise UserConfigValidationException("Desired class cannot be opposite if the number of classes is more than 2.")
+        elif isinstance(desired_class_input, int):
             if desired_class_input >= 0 and desired_class_input < num_output_nodes:
                 target_class = desired_class_input
+                return target_class
             else:
-                raise ValueError("Desired class should be within 0 and num_classes-1.")
-        return target_class
+                raise UserConfigValidationException("Desired class not present in training data!")
+        else:
+            raise UserConfigValidationException("The target class for {0} could not be identified".format(
+                                                desired_class_input))
 
     def infer_target_cfs_range(self, desired_range_input):
         target_range = None
@@ -485,15 +503,15 @@ class ExplainerBase:
         if hasattr(model_score, "shape") and len(model_score.shape) > correct_dim:
             model_score = model_score[0]
         # Converting target_cf_class to a scalar (tf/torch have it as (1,1) shape)
-        target_cf_class = self.target_cf_class
-        if hasattr(self.target_cf_class, "shape"):
-            if len(self.target_cf_class.shape) == 1:
-                target_cf_class = self.target_cf_class[0]
-            elif len(self.target_cf_class.shape) == 2:
-                target_cf_class = self.target_cf_class[0][0]
-        target_cf_class = int(target_cf_class)
+        if self.model.model_type == 'classifier':
+            target_cf_class = self.target_cf_class
+            if hasattr(self.target_cf_class, "shape"):
+                if len(self.target_cf_class.shape) == 1:
+                    target_cf_class = self.target_cf_class[0]
+                elif len(self.target_cf_class.shape) == 2:
+                    target_cf_class = self.target_cf_class[0][0]
+            target_cf_class = int(target_cf_class)
 
-        if self.model.model_type == "classifier":
             if self.num_output_nodes == 1:  # for tensorflow/pytorch models
                 pred_1 = model_score[0]
                 validity = True if ((target_cf_class == 0 and pred_1<= self.stopping_threshold) or (target_cf_class == 1 and pred_1>= self.stopping_threshold)) else False
@@ -572,3 +590,11 @@ class ExplainerBase:
             KD_tree = KDTree(dummies)
 
         return dataset_with_predictions, KD_tree, predictions
+
+    def round_to_precision(self):
+        # to display the values with the same precision as the original data
+        precisions = self.data_interface.get_decimal_precisions()
+        for ix, feature in enumerate(self.data_interface.continuous_feature_names):
+            self.final_cfs_df[feature] = self.final_cfs_df[feature].astype(float).round(precisions[ix])
+            if self.final_cfs_df_sparse is not None:
+                self.final_cfs_df_sparse[feature] = self.final_cfs_df_sparse[feature].astype(float).round(precisions[ix])
